from .lrng import LabeledRange, LabeledRanges
from .numba import coalesce, label_range, relevant_labels
name = 'lrng'

major = 0
minor = 0
patch = 14

version = '{}.{}.{}'.format(major, minor, patch)
description = 'Labeled Ranges'
